/**
 * QBodyGraph
 * by Daniel Steinbock <steinbock@stanford.edu>
 * 
 * Multiple graph visualizations for QMovement body movement data
 *
 */ 

import toxi.math.*;

class QBodyGraph {
  String name;
  int number;
  int[] data;
  String[] frame;
  float[] time;
  PGraphics buf;
  PImage bodyGraph;
  int myGraphHeight;

  // Color Graph
  int globalMax = 0;
  int[] movingMax;
  int maxWindow = 0;
  
  // Line Graph
  int[] movingAvg;
  int avgWindow = 0;
  ScaleMap logMap;
  
  // Step Graph
  byte[] stepGraph;
  int minMove;    // threshold at which individual movement gets included in combined stepGraph
  int skipBody = -1;
  
  int visibleStart;
  
  QBodyGraph(String theName, int dataSize, int theNumber, int theGraphHeight) {
    name = theName;
    number = theNumber;
    myGraphHeight = theGraphHeight;
    data = new int[ dataSize ];
    frame = new String[ dataSize ];
    time = new float[ dataSize ];
    bodyGraph = createImage( dataSize, myGraphHeight, RGB );
    movingMax = new int[ dataSize ];
    maxWindow = 0;
    movingAvg = new int[ dataSize ];
    avgWindow = 42;
    logMap = new ScaleMap(0,1,myGraphHeight-1,0);
    logMap.setMapFunction(new ExponentialInterpolation());
    buf = createGraphics( dataSize, myGraphHeight, JAVA2D );
    visibleStart = 0;
    stepGraph = new byte[ dataSize ];
    minMove = 50;
  }
  
  QBodyGraph() {
    // empty
  }

  void setData(int count, int dataVal) {
    data[ count ] = dataVal;
    if( dataVal > globalMax ) globalMax = dataVal;
  }
  
  void setFrame(int count, String frameVal) {
    frame[ count ] = frameVal;
  }

  void setTime(int count, float timeVal) {
    time[ count ] = timeVal;
  }
  
  void generate() {
    if( getMode() == GRAPH_MODE )
      generateAvgGraph();
    else if( getMode() == COLOR_MODE )
      generateColorGraph();
    else if( getMode() == ANALYSIS_MODE )
      generateStepGraph();
  }
  
  void refresh() {   // don't regenerate data, just grab a new part of the already generated buffer
    bodyGraph = buf.get(visibleStart, 0, (int)(width * zoomFactor), myGraphHeight-1);
  }

  void generateStepGraph() {
    computeMovingAverage();
    buf.beginDraw();
    buf.background(graphColorBG); 
    buf.fill(graphColor);
    buf.beginShape();
    buf.stroke(graphColorStroke);
    buf.noStroke();
    buf.vertex(0, myGraphHeight-1);        // starting point
    for(int i = 0; i<data.length;++i){
    //  if( movingAvg[i] >= minMove ){             // making minMove a filter applied before averaging is done, by commenting this out
      if( data[i] >= minMove ){
        buf.vertex(i-1, 2);     // make square graph
        buf.vertex(i, 2);
        stepGraph[i] = 1;
      }
      else {
        buf.vertex(i-1, myGraphHeight-1);    // make square graph
        buf.vertex(i,myGraphHeight-1);
        stepGraph[i] = 0;
      }
    }
    buf.vertex(data.length-1, myGraphHeight-1); // ending point
    buf.endShape();
    buf.endDraw();
    refresh();
  }
  
  void generateAvgGraph() {
    computeMovingAverage();
    buf.beginDraw();
    buf.background(graphColorBG);
    buf.noFill();
    buf.beginShape();
    buf.stroke(graphColor);
    buf.vertex(0, myGraphHeight-1);        // starting point
    logMap.setInputRange(0,log(globalMax));
    for(int i = 0; i<data.length;++i){
      if( movingAvg[i] != 0 && movingAvg[i] >= minMove )
        buf.vertex(i,(float)logMap.getMappedValueFor(log(movingAvg[i])));
      else
        buf.vertex(i,myGraphHeight-1);
    }
    buf.vertex(data.length-1, myGraphHeight-1); // ending point
    buf.endShape();
    buf.endDraw();
    refresh();
  }
  
  void computeMovingAverage() {
    int start = 0;
    int end = 0;

    if(avgWindow == 0) {                    // use raw data if no moving average window
      movingAvg = Arrays.copyOf( data, data.length );
    }
    else {
      for(int i = 0; i < data.length; ++i ){
        start = i - avgWindow;
        end = i + avgWindow;
        if(i < avgWindow)                            // if point i is one of the left-most
          start = 0;          
        if(i > data.length - 1 - avgWindow)        // if this point i is one of the right-most
          end = data.length - 1;
        movingAvg[i] = int(avg( Arrays.copyOfRange( data, start, end )));
      }
    }
  }

  void randomize() {
    Vector rando = autobox( data );
    Collections.shuffle( rando );
    data = unboxInt( rando );
  }

  Vector autobox(int[] g) {
    Vector v = new Vector(g.length);
    for(int i=0; i<g.length; ++i) {
      v.add(new Integer(g[i]));
    }
    return v;
  }

  int[] unboxInt( Vector v ) {
    int[] ints = new int[ v.size() ];
    for(int i=0; i<v.size(); ++i) {
      ints[i] = ((Integer)v.get(i)).intValue();
    }
    return ints;
  }

  float avg(int[] vals) {
    float result = 0;
    for(int i=0; i<vals.length; ++i) {
      result += vals[i];
    }
    return result / vals.length;
  }
  
  void computeMovingMax() {
    int start = 0;
    int end = 0;

    if(maxWindow == 0) {                    // use globalMax, otherwise calculate movingMax
      Arrays.fill(movingMax, globalMax);
    }
    else {
      for(int i = 0; i < data.length; ++i ){
        start = i - maxWindow;
        end = i + maxWindow;
        if(i < maxWindow)                            // if point i is one of the left-most
          start = 0;          
        if(i > data.length - (maxWindow + 1))        // if this point i is one of the right-most
          end = data.length - 1;
        movingMax[i] = max( Arrays.copyOfRange( data, start, end));
      }
    }
  }
  
  void generateColorGraph() {
    color c;
    
    computeMovingMax();
    bodyGraph.loadPixels();
    for(int x=0; x < data.length; ++x ){
      if( data[x] > 0 )
//        c = color(255, 255, map( data[x], 1, dataMax, 64, 255));            // black background
        c = color(map( data[x], 1, movingMax[x], 240, 0), map( data[x], 1, movingMax[x], 240, 0), 255);    // white background
      else
        c = color(255,255,255);
      for(int h=0; h < myGraphHeight; ++h ){
        bodyGraph.pixels[h*bodyGraph.width + x] = c;
      }
    }
    bodyGraph.updatePixels();
  }
  
  PImage getBodyGraph() {
    return bodyGraph;
  }

  void setMinMove(int val) {
    if( val > 0 )
      minMove = val;
    else
      minMove = 1;
  }
  
  int getMinMove() {
    if( minMove == 1 )
      return 0;
    else
      return minMove; 
  }

  void setAvgWindow(int window) {
    if( window >= 0 && window < data.length / 2 )
      avgWindow = window;
  }
  
  int getAvgWindow() {
    return avgWindow;
  }
  
  void setMaxWindow(int window) {
    if( window >= 0 && window < data.length / 2 )
      maxWindow = window;
  }
  
  int getMaxWindow() {
    return maxWindow;
  }
  
  int getVisibleStart() {
    return visibleStart;
  }
  
  void setVisibleStart(int jump) {
    if( jump < 0 )                // hitting left boundary, show a little black to show it's the end
      visibleStart = 0;
    if( jump > (data.length - int(width*zoomFactor)) )    // hitting right boundary, show a little black to show it's the end
      visibleStart = data.length - int(width*zoomFactor);
    else
      visibleStart = jump;
  }
  
  void incrementSkipBody() {      
    skipBody++;
    if(skipBody >= numBodies)
      skipBody = -1;
    println("Skipping Body "+ skipBody);
  }
}
