/**
 * QBodyStepGraph
 * by Daniel Steinbock <steinbock@stanford.edu>
 * 
 * Step function graph visualization for QMovement body movement data analysis
 *
 */ 

class QBodyStepGraph extends QBodyGraph {
  
  int numSteps;
  float stepSize;
  int stepWindow;
  byte[] unionBin;
  Vector cleanGraph;
  StringBuffer cleanStringBuffer;
  // printable output vars
  PGraphics outputBuffer;
  PImage[] outputImage;
  int numRows;
  int numPages;
  int outputMargin = 25;
  int outputImageHeight = 1840;
  int outputImageWidth = 2700 + outputMargin;
  int outputRowWidth = outputImageWidth - outputMargin;
  
  int myGraphHeightPrint = 300;
  boolean ministryFlag = true;

  
  QBodyStepGraph(String theName, int dataSize, int bodyCount, int theGraphHeight) {
    super(theName, dataSize, bodyCount, theGraphHeight);
    numSteps = bodyCount;
    stepSize = round((float)(myGraphHeight - 3) / numSteps);      // -3 to make the max y coordinate 2
    stepWindow = 45;
    unionBin = new byte[ bodyCount ];
    cleanGraph = new Vector(stepGraph.length / 30);
  }
  
  void generateStepGraph() {
    computeMovingUnion();      // compute and fill stepGraph[]
    buf.beginDraw();
    buf.background(graphColorBG);
//    // draw grid
//    buf.stroke(tickColor);
//    for(int i=1; i <= numBodies; ++i ){
//      buf.line(0, myGraphHeight - i * stepSize - 1, buf.width-1, myGraphHeight - i * stepSize - 1);
//    }  
    // draw step graph
    buf.fill(graphColorMinistry);
    buf.beginShape();
    buf.stroke(graphColorStroke);  // 100,200,240
    buf.noStroke();
    buf.vertex(0, myGraphHeight-1);
    for(int i = 0; i<data.length;++i){
      if( frame[i].endsWith("v") != ministryFlag ){
        buf.vertex(i, myGraphHeight-1);
        buf.endShape();
        if( frame[i].endsWith("v") )
          buf.fill(graphColorMinistry);
        else
          buf.fill(graphColor);
        buf.beginShape();
        buf.vertex(i, myGraphHeight-1);
        ministryFlag = frame[i].endsWith("v");
      }
      if( stepGraph[i] > 0 ){
        buf.vertex(i-1, myGraphHeight-1 - stepSize * (stepGraph[i]));
        buf.vertex(i, myGraphHeight-1 - stepSize * (stepGraph[i]));
      }
      else {
        buf.vertex(i-1, myGraphHeight - 1);
        buf.vertex(i, myGraphHeight - 1);
      }
    }
    buf.vertex(data.length-1, myGraphHeight-1);
    buf.endShape();
    buf.endDraw();    
    drawTimestamps();
    bodyGraph = buf.get(visibleStart, 0, width, myGraphHeight-1);    
  }
  
  void computeMovingUnion() {
    int start = 0;
    int end = 0;

    for(int i = 0; i < data.length; ++i ){
      stepGraph[i] = 0;
      start = i - stepWindow;
      end = i + stepWindow;
      if(i < stepWindow)                            // if point i is one of the left-most
        start = 0;          
      if(i > data.length - 1 - stepWindow)        // if this point i is one of the right-most
        end = data.length - 1;
      for(int x = start; x <= end; ++x ){                // for each frame within the averaging-window
        for(int body = 0; body < numBodies; ++body ){    // for each body
          if(body != skipBody)
            unionBin[body] += bodies[body].stepGraph[x];   // add it to the bin if its positive
        }
      }
      for(int u = 0; u < unionBin.length; ++u) {        // if any body was positive in the averaging-window, raise the stepgraph
        if( unionBin[u] > 0 )
          stepGraph[i] += 1;
        unionBin[u] = 0;                                // initialize for next time
      }
    }
  }
  
  int getStepWindow() {
    return stepWindow;
  }
  
  void setStepWindow(int window) {
    if( window >= 0 && window < data.length / 2 )
      stepWindow = window;
  }

  Vector autobox(byte[] g) {
    Vector v = new Vector(g.length);
    for(int i=0; i<g.length; ++i) {
      v.add(new Byte(g[i]));
    }
    return v;
  }
  
  Vector autobox(char[] g) {
    Vector v = new Vector(g.length);
    for(int i=0; i<g.length; ++i) {
      v.add(new Character(g[i]));
    }
    return v;
  }

  StringBuffer unbox(Vector v) {
    char[] chars = new char[ v.size() ];
    for(int i=0; i<v.size(); ++i) {
      chars[i] = ((Character)v.get(i)).charValue();
    }
    return new StringBuffer( new String( chars ));
  }

  void outputStepData() {
    QDataOutput q = new QDataOutput("StepSize,", 1);
    q.writeData(cleanGraph);
    q.finish();
  }

  void computeHistogram() {
    byte c = 0;
    byte lastPeak = 0;
    int[] histogram = new int[ numBodies + 1];
    
    for(int i=0;i < cleanGraph.size(); ++i) {
      c = ((Byte)cleanGraph.get(i)).byteValue();
      if( c == 0 && lastPeak != 0 ){
        histogram[ 0 ] += 1;       // count silences
        histogram[ lastPeak ] += 1;  // count events of each size
        lastPeak = 0;
      }
      else if( c > lastPeak)
        lastPeak = c;
    }
    if( lastPeak != 0)
      histogram[ lastPeak ] += 1;  // count final event
    
    //print result
    println("HISTOGRAM");
    for(int h=0; h<histogram.length; ++h ){
      println(h + "\t" + histogram[ h ]);
    }
    
    //print probabilities
    println("PROBABILITIES");
    int total = histogram[0] + 1;
    println("P = " + histogram[1] + " / " + total + " = " + (float(histogram[1]) / total));
    println("Q = " + (total - histogram[1])  + " / " + total + " = " + (float(total - histogram[1]) / total));

  }

  void computeQProbabilities() {
    Vector stepData = autobox(stepGraph);
    Vector bootstrapData = autobox(stepGraph);
    cleanRedundancies(stepData);  // populates the cleanGraph
    println("Sample Probabilities:");
    computeProbabilities();    // calculates and prints probabilities from cleanGraph
    computeHistogram();


//    outputStepData();

        
//    cleanRedundancies(bootstrapData);
//    cleanGraph = autobox(cleanStringBuffer.toString().toCharArray());
//    Collections.shuffle(cleanGraph);
//    cleanStringBuffer = unbox(cleanGraph);
//    println("Bootstrap Probabilities:");
//    computeProbabilities();
  }

  void cleanRedundancies(Vector v) {
    byte last = -1;
    byte current = 0;
    cleanGraph.clear();
    cleanStringBuffer = new StringBuffer(int(v.size() / 10));
    for(int i=0; i<v.size(); ++i) {
      current = (Byte)v.get(i);
      if( current != last ){          
        cleanGraph.add(v.get(i));
        last = current;
        cleanStringBuffer.append( Integer.toHexString( ((Byte)v.get(i)).intValue() ));
      }
    }
    println("clean redundancies: " + stepGraph.length + " -> " + cleanGraph.size());
  }

  void computeProbabilities() {
    int noContingencyCounter = 0;
    int contingencyCounter = 0;
    int total = 0;
    
    String cleanString = cleanStringBuffer.toString();
    noContingencyCounter = cleanString.split("010").length-1;
    contingencyCounter = cleanString.split("012").length-1;
    contingencyCounter += cleanString.split("013").length-1;
    contingencyCounter += cleanString.split("020").length-1;
    contingencyCounter += cleanString.split("021").length-1;
    contingencyCounter += cleanString.split("023").length-1;
    
//    byte a = (Byte)cleanGraph.get(0);
//    byte b = (Byte)cleanGraph.get(1);
//    byte c = 0;
//    for(int i=2;i < cleanGraph.size(); ++i) {
//      c = (Byte)cleanGraph.get(i);
//      if( a==0 && b==1 && c==0 )      // no contingency
//        noContingencyCounter++;
//      else if( a==0 && b==1 && c==2 )  // yes contingency
//        contingencyCounter++;
//      a = b;
//      b = c;
//    }
    total = noContingencyCounter + contingencyCounter;
    println("P = " + noContingencyCounter + " / " + total + " = " + (float(noContingencyCounter) / total));
    println("Q = " + contingencyCounter + " / " + total + " = " + (float(contingencyCounter) / total));
  }

  void drawPrintable() {
    bigGraphHeight = myGraphHeightPrint;
    float stepSizePrint = (float)(myGraphHeightPrint - 3) / numSteps;      // -3 to make the max y coordinate 2
    computeMovingUnion();      // compute and fill stepGraph[]
    buf.beginDraw();
    buf.background(graphColorBGPrint);
    buf.fill(graphColorPrint);
    buf.beginShape();
    buf.stroke(graphColorStrokePrint);  // 100,200,240
    buf.vertex(0, myGraphHeightPrint-1);
    for(int i = 0; i<data.length;++i){
      if( stepGraph[i] > 0 ){
        buf.vertex(i-1, myGraphHeightPrint-1 - stepSizePrint * (stepGraph[i]));
        buf.vertex(i, myGraphHeightPrint-1 - stepSizePrint * (stepGraph[i]));
      }
      else {
        buf.vertex(i-1, myGraphHeightPrint - 1);
        buf.vertex(i, myGraphHeightPrint - 1);
      }
    }
    buf.vertex(data.length-1, myGraphHeightPrint-1);
    buf.endShape();
    
    //draw timestamps
    buf.endDraw();    
    drawTimestamps();
    bodyGraph = buf.get(visibleStart, 0, width, myGraphHeightPrint-1);    
  }
  
  void outputGraph() {
    // ** 1. As a first step, re-draw the output graph with a height such that step size is equal to one second
    // 2. Auto-split the graph into dimensions that print well on letter paper in landscape (a whole session should fit on 2 pages)
    // width = 2700 px, height = 11 rows * 153 px = 1683 px        -> 2700 x 1683
    // 3. Repeat row numbers and grid for each graph wrap
    
    // Generate custom-sized output graph w/ 16 pixel step-size
    myGraphHeight = 16 * numBodies;
    stepSize = 16;
    bodyGraph = createImage( data.length, myGraphHeight, RGB );
    buf = createGraphics( data.length, myGraphHeight, JAVA2D );
    generate();
    // create output buffer, initialize white, then copy the whole graph buffer to it
    outputBuffer = createGraphics(buf.width, buf.height + outputMargin, JAVA2D);
    if( outputBuffer.width < outputRowWidth )
      outputBuffer = createGraphics(outputRowWidth, buf.height + outputMargin, JAVA2D);
    initWhite(outputBuffer);
    outputBuffer.set(0, 1, buf.get());
    // draw the chrome (grid, axis labels, timestamps)
    outputGraphGrid();
//    outputGraphTimestamps();
    // cut up buffer into rows for printing (currently has top margin for timestamps but no side margin yet for axis
    cutupOutputGraph();
    // TODO: 1. move timestamps to below the graph, with horizontal line
    // 2. properly support fig_1 and tail ends of data, which don't have full window width

    // write to file
//    outputBuffer.save(graphFilename);
    for(int p=0; p<numPages; ++p){
      String graphFilename = path.substring(0,path.length()-4) + "_" + (stepWindow / 15) + "s_" + int(zoomFactor) + "x-" + p + ".tif";
      println("Outputting " + graphFilename);
      outputImage[p].save(graphFilename);
    }
  }

  void outputGraphGrid() {
    // draw horizontal grid lines
    outputBuffer.stroke(graphColorBG);
    for(int i=1; i <= numBodies; ++i ){
      outputBuffer.line(0,
                       myGraphHeight - i * stepSize, 
                       buf.width-1,
                       myGraphHeight - i * stepSize);
    }
  }
  
  void outputGraphAxis(PGraphics img) {
    // draw axis labels
    img.textFont(printFont, 8.0);
    img.stroke(tickColor);
    img.fill(textColor);
    for(int i=1; i <= numBodies; ++i ){
      img.line(0, myGraphHeight - i * stepSize, 
                        8, myGraphHeight - i * stepSize);
      img.text(i, 1, 12 + myGraphHeight - i * stepSize);
    } 
  }
  
  void outputGraphTimestamps(PGraphics img, int startTime) {
    float timestamp;
    img.textFont(printFont, 16.0);
    img.fill(textColor);
    img.stroke(tickColor);
    for(int i=0; i<buf.width; ++i ){
      timestamp = i;
      if( timestamp % (225*zoomFactor) == 0 )
        img.text((timestamp+startTime)/900 + "m", int((i+2)/zoomFactor) + outputMargin, img.height - 10);
      if( timestamp % (75*zoomFactor) == 0 )
        img.line(int(i/zoomFactor) + outputMargin, outputBuffer.height - outputMargin + 1, int(i/zoomFactor) + outputMargin, outputBuffer.height - 10);  
    }
  }

  void cutupOutputGraph() {
    // calculate how many rows and pages needed to fit this graph
    int outputRowHeight = myGraphHeight + outputMargin;
    int outputRowsPerPage = floor((float)outputImageHeight / outputRowHeight);
    numRows = ceil((float)outputBuffer.width / (outputRowWidth * zoomFactor));
    numPages = ceil((float)numRows / outputRowsPerPage);
    PGraphics rowBuffer = createGraphics(outputImageWidth, outputRowHeight, JAVA2D);
    
//    println("rowheight = " + outputRowHeight);
//    println("row-width = " + outputRowWidth);
//    println("rows-per = " + outputRowsPerPage);
    println("rows = " + numRows);
    println("pages = " + numPages);
    
    outputImage = new PImage[ numPages ];
    int row = 0;
    for(int p=0; p<numPages; ++p){
      println("creating page " + p);
      outputImage[p] = createImage(outputImageWidth, outputImageHeight, RGB);
      initWhite( outputImage[p] );
      // loop over outputBuffer, grabbing pieces and adding them to final image
      for(int r=0; r<outputRowsPerPage; ++r){
        initWhite(rowBuffer);
        println("grabbing " + row*outputRowWidth*zoomFactor + " - " + (row+1)*outputRowWidth*zoomFactor);
        rowBuffer.set(outputMargin, 0, outputBuffer.get(int(row*outputRowWidth*zoomFactor), 0, int((row+1)*outputRowWidth*zoomFactor), outputRowHeight));

// commenting out because it's buggy -- zoom not working
//        rowBuffer.image(outputBuffer.get(int(row*outputRowWidth*zoomFactor), 0, int((row+1)*outputRowWidth*zoomFactor), outputRowHeight),
//                        outputMargin, 0, outputRowWidth, outputRowHeight);
        outputGraphAxis( rowBuffer );
        outputGraphTimestamps( rowBuffer, int((p*outputRowsPerPage*outputRowWidth*zoomFactor) + r*outputRowWidth*zoomFactor) );
        outputImage[p].set(0, r*outputRowHeight, rowBuffer.get());
        if( ++row >= numRows )
          break;
      }
    }
    

    
  }
  
  void initWhite(PImage img) {
    img.loadPixels();
    for (int i = 0; i < img.pixels.length; i++) {
      img.pixels[i] = color(255,255,255); 
    }
    img.updatePixels();
  }
}
