/**
 * QColorGraph
 * by Daniel Steinbock <steinbock@stanford.edu>
 * 
 * Writes movement time series data to a file, rendered as a heatmap color bar.
 *
 */ 

/* PSEUDOCODE v. 1.0

  Ask user for a CSV file to load
  Read in CSV file
  Parse it into data structures:
    - one array of ints for each body
    - one array for the total
    - convert these to colors according to color palette
  Draw()
    - update the visible graph PImages (for current zoom level):
    for each body PImage
      for each column in the currently visible range (tmin to tmin + twidth)
        get C (movement data point, now converted to a Color)
          - if zoom level = 1, return C
          - if zoom level = 2, return blend of 2 neighbors
          - if zoom level = 4, return blend of 4 neighbors
          etc.
        update this body's PImage for this column w/ C
  Swipe()
    - update tmin, add or subtract
  Zoom()
    - columns = 1152
    - map frames to columns
    - update twidth
      zoom = 1: twidth = 1152  (1m16s)
      zoom = 2: twidth = 2304  (2m32s)
      zoom = 4: twidth = 4608  (~5m)
      etc.
  Save()
    - Ask user: Save current viewport or entire dataset?
    - output to vector-friendly format (pdf, gif?)
    
    
  TO DO:
  - local maxima slider
  - produce statistical rows: correlation between bodies
  - use heat map colors
  - scrolling
  - zooming

}
*/

QBodyGraph[] bodies;
QBodyStepGraph total;

int numBodies = 0;
int graphHeight = 0;
int graphSpacing = 1;
int bigGraphHeight = 300;
int bigGraphSpacing = 1;
float zoomFactor = 1.0;
final static int COLOR_MODE = 0;
final static int GRAPH_MODE = 1;
final static int ANALYSIS_MODE = 2;
int mode = ANALYSIS_MODE;

color bgColor = color(60);
color graphColor = color(50,100,120);
color graphColorBG = color(0,0,30);
color graphColorStroke = color(100,200,240);
color graphColorMinistry = color(50,120,100);
color locationBarColor = color(240,200,100);
color locationBarColorBG = color(60);
color textColor = color(210,210,210);
color tickColor = color(200);

color graphColorPrint = color(50,120,140);
color graphColorBGPrint = color(255);
color graphColorStrokePrint = graphColorPrint;
color graphColorMinistryPrint = color(50,140,100);
color textColorPrint = color(0);
color tickColorPrint = color(160);
color bgColorPrint = color(255);
color locationBarColorBGPrint = tickColorPrint;
color locationBarColorPrint = tickColorPrint;

//final color bgColor = color(180);
//final color graphColor = color(0);
//final color graphColorBG = color(255);
//final color graphColorStroke = color(160);
//final color locationBarColor = color(255);
//final color locationBarColorBG = color(255);
//final color textColor = color(45);

String path; // input filename
PFont screenFont;
PFont printFont;

void setup() {
  size(1400, 800);
  background(bgColor);
  strokeWeight(1);
  noSmooth();
  hint(ENABLE_NATIVE_FONTS);
  
  // select path and load CSV file
  path = selectInput("Select a QMovement CSV data file ...");
//  screenFont = loadFont("LucidaGrande-12.vlw");
//  screenFont = loadFont("CourierNewPSMT-14.vlw");
  screenFont = createFont("pixelmix", 8.0, false);
  textFont(screenFont);
  printFont = loadFont("PFRondaSeven-16.vlw");
  String[] lines = loadStrings(path);
  String[] headers = splitTokens(lines[0], ",");
  numBodies = headers.length - 4;
//
//  bigGraphHeight = graphHeight * numBodies;
//  bigGraphHeight + 
//  graphHeight = (height - bigGraphHeight) / numBodies;
//  
//  bigGraphHeight = (height - bigGraphSpacing) / 2;
  graphHeight = (height - bigGraphHeight - bigGraphSpacing) / numBodies - graphSpacing;

  println("num bodies = " + numBodies);
  bodies = new QBodyGraph[ numBodies ];
  for(int body = 0; body < numBodies; ++body ){
    bodies[body] = new QBodyGraph( headers[4 + body], lines.length - 2, body, graphHeight );
  }
  total = new QBodyStepGraph( "total", lines.length - 2, numBodies, bigGraphHeight );
  // read graph data
  for(int i = 1; i < lines.length - 1; i++){        // read CSV file except for header and ceiling lines
    String[] values = splitTokens(lines[i], ",");
    for(int body = 0; body < numBodies; ++body ){
      bodies[body].setFrame(int(values[0]), values[1]);
      bodies[body].setTime(int(values[0]), float(values[2]));
      bodies[body].setData(int(values[0]), int(values[4 + body]));
    }
    // total body
    total.setFrame(int(values[0]), values[1]);
    total.setTime(int(values[0]), float(values[2]));
    total.setData(int(values[0]), int(values[3]));
  }
  // generate graphs
  regenerate();
  
  // temporary batch mode to produce printables for dissertation
  setAnalysisMode();
//  drawPrintable();
//  total.outputGraph();
//  quit();
}

void regenerate() {
  // generate graphs
  for(int body = 0; body < numBodies; ++body ){ 
    bodies[body].generate();
  }
  total.generate();
}

void refresh() {
  for(int body = 0; body < numBodies; ++body ){ 
    bodies[body].refresh();
  }
  total.refresh();  
}

void draw() {
  PImage graph;
  PImage totalGraph = total.getBodyGraph();

  image(totalGraph, 0, 0, width, bigGraphHeight);
  int y = bigGraphHeight + bigGraphSpacing;                // start out below the totalGraph
  for(int body = 0; body < numBodies; ++body ){
    graph = bodies[body].getBodyGraph();
    image(graph, 0, y, width, graphHeight);
    y += graphHeight + graphSpacing;
  }
  drawTimestamps();
  drawLocationBar();
  drawVerticalAxis();
}

void drawTimestamps() {  // intended for top graph only
  float timestamp;
  fill(textColor);
  stroke(tickColor);
  for(int i=1; i<width; ++i ){
    timestamp = total.visibleStart + i * zoomFactor;
    if( timestamp % (225*zoomFactor) == 0 ){
      text(timestamp/900 + "m", i+2, 15);
      line(i,0,i,20);  
    }
  }
}

void drawLocationBar() {
  stroke(locationBarColorBG);
  float theWidth = (width*zoomFactor / float(total.data.length)) * width;
  float x = (total.getVisibleStart() / float(total.data.length)) * width;
  fill(locationBarColorBG); // 60
  rect(0, bigGraphHeight, width, bigGraphSpacing);    // the bar
  stroke(locationBarColor);
  fill(locationBarColor);   // 240,200,100
  rect(x, bigGraphHeight, theWidth, bigGraphSpacing);  // the location marker
  noFill();
}

void drawVerticalAxis() {
  // draw horizontal grid lines
  if( mode == ANALYSIS_MODE ){
    stroke(graphColorBG);
    for(int i=1; i <= numBodies; ++i ){
      line(0, bigGraphHeight - i * total.stepSize - 1, width-1, bigGraphHeight - i * total.stepSize - 1);
    }
  }
  // draw axis labels
  stroke(tickColor);
  fill(textColor);
  for(int i=1; i <= numBodies; ++i ){
    line(0, bigGraphHeight - i * total.stepSize - 1, 8, bigGraphHeight - i * total.stepSize - 1);
    text(i, 1, 10 + bigGraphHeight - i * total.stepSize);
  }
}

void keyPressed() {
  if (key == CODED) {
    switch(mode) {
      case(COLOR_MODE):
        if(keyCode == UP)
          incrementMaxWindow(75);
        else if(keyCode == DOWN)
          incrementMaxWindow(-75);
        break;
      case(GRAPH_MODE):
      case(ANALYSIS_MODE):
        switch(keyCode) {
          case(UP):
            incrementAvgWindow(1);
            break;
          case(DOWN):
            incrementAvgWindow(-1);
            break;
          case(RIGHT):
            shiftVisible(int(width/10*zoomFactor));
            break;
          case(LEFT):
            shiftVisible(-int(width/10*zoomFactor));
            break;
        }
        break;
      default:
        println("ERROR. Unknown mode!");
        quit();
        break;
    }
  }
  else if (key == 'h') {
//    bigGraphHeight = 600;
    setAnalysisMode();
//    total.drawPrintable();
    drawPrintable();
  }
  else if (key == 'p') {
    setAnalysisMode();
    total.computeQProbabilities();
//    computeBootstrapProbabilites();
//    total.computeQProbabilities();
  }
  else if (key == 'g') {
    setGraphMode();
  }
  else if (key == 'c') {
    setColorMode();
  }
  else if (key == 's') {
    setAnalysisMode();
  }
  else if (key == 'o') {
    total.outputGraph();
  }
  else if (key == 'f') {
    incrementStepWindow(15);
  }
  else if (key == 'v') {
    incrementStepWindow(-15);
  }
  else if (key == 'a') {
    incrementZoomFactor(0.5);
  }
  else if (key == 'z') {
    if( zoomFactor*width < total.data.length ) incrementZoomFactor(2.0);
  }
  else if (key == '0') {
    resetMaxWindow();
  }
  else if (key == '+') {
    incrementMinMove(50);    
  }
  else if (key == '-') {
    incrementMinMove(-50);
  }
  else if (key == '1') {
    total.incrementSkipBody();
  }
  else if (key == 'q') {  // Quit and flush output buffer
    quit();
  }
}

void computeBootstrapProbabilites() {
  for(int body = 0; body < numBodies; ++body ){
    // randomize body's movement data
    bodies[body].randomize();
  }
  regenerate();
}

void drawPrintable() {
  background(bgColorPrint);
  graphColor = graphColorPrint;
  graphColorBG = graphColorBGPrint;
  graphColorStroke = graphColorStrokePrint;
  graphColorMinistry = graphColorMinistryPrint;
  locationBarColor = locationBarColorPrint;
  locationBarColorBG = locationBarColorBGPrint;
  textColor = textColorPrint;
  tickColor = tickColorPrint;
  regenerate();
}

void setMode(int newMode) {
  mode = newMode;
}

int getMode() {
  return mode;
}

void setGraphMode() {
  setMode(GRAPH_MODE);
  background(bgColor);
  regenerate();
  println("Graph mode");
}

void setColorMode() {
  setMode(COLOR_MODE);
  background(bgColor);
  regenerate();
  println("Color mode");
}

void setAnalysisMode() {
  setMode(ANALYSIS_MODE);
  background(bgColor);
  regenerate();
  println("Analysis mode");
}

void printTimestamp() {
  println("TIME RANGE SHOWN: " + nf(total.time[ total.getVisibleStart() ],0,2) + "  -  " + nf(total.time[ total.getVisibleStart() + width - 1],0,2));
}

void shiftVisible(int val) {
  for(int body = 0; body < numBodies; ++body ){
    bodies[body].setVisibleStart( bodies[body].getVisibleStart() + val);
    bodies[body].refresh();
  }
  total.setVisibleStart( total.getVisibleStart() + val);
  total.refresh();
//  printTimestamp();
}

void incrementMinMove(int val) {
  for(int body = 0; body < numBodies; ++body ){
    bodies[body].setMinMove( bodies[body].getMinMove() + val );
  }
  total.setMinMove( total.getMinMove() + val );
  regenerate();
  println("minMove = " + bodies[0].getMinMove() );
}

void incrementStepWindow(int val) {
  total.setStepWindow( total.getStepWindow() + val );
  total.generate();
  println("stepWindow = +/- " + (total.getStepWindow() / 15) + " seconds" );
}

void incrementAvgWindow(int val) {
  for(int body = 0; body < numBodies; ++body ){
    bodies[body].setAvgWindow( bodies[body].getAvgWindow() + val );
  }
  total.setAvgWindow( total.getAvgWindow() + val );
  regenerate();
  println("avgWindow = " + bodies[0].getAvgWindow() );
}

void resetMaxWindow() {
  for(int body = 0; body < numBodies; ++body ){
    bodies[body].setMaxWindow( 0 );
  }
  total.setMaxWindow( 0 );
  regenerate();
  println("maxWindow = global max");
}

void incrementMaxWindow(int val) {
  for(int body = 0; body < numBodies; ++body ){
    bodies[body].setMaxWindow( bodies[body].getMaxWindow() + val );
  }
  total.setMaxWindow( total.getMaxWindow() + val );
  regenerate();
  println("maxWindow = " + int(bodies[0].getMaxWindow() / 15 * 2) + " / " + int(total.data.length / 15));
}

void incrementZoomFactor(float val) {
  zoomFactor = zoomFactor * val;
  println("zoom factor = " + zoomFactor);
  refresh();
}

void quit() {
//  output.finish();
  println("Quitter");
  exit(); // Stop the program  
}




//  long movieLength;
//  final static int avgWindowSize = 15;      // number of frames to blend together for each pixel line in the graph (30 = 1 second)
//  int[] dataBuffer;
//  final String colorGraphFilename = "movement_data.jpg";
//  PGraphics colorGraphBuffer;
  
//  QColorGraph(int numBodies, long numFrames) {
//    dataBuffer = new int[avgWindowSize];        // create data buffer for averaging data
//    colorGraphBuffer = createGraphics((int)numFrames, 40, JAVA2D);


//    fileOutput = createWriter(colorGraphFilename);
//
//    fileOutput.print("Frame,Time,Total,");
//    fileOutput.print("\n");
    
//  }

//  void writeData(int currFrame, int bodyNum, int diffData) {
//    dataBuffer[currFrame % avgWindowSize] = diffData;
//    fileOutput.print(currFrame+"," + movie.time() + "," + qdata.getMovementTotal() + ",");
//    fileOutput.print("\n");
//  }


//  public void finish(){
//    fileOutput.flush(); // Write the remaining data
//    fileOutput.close(); // Finish the file
//  }

