/**
 * QData
 * by Daniel Steinbock <steinbock@stanford.edu>
 * 
 * Data structure for calculating movement data for QMovement
 *
 */ 

class QData {
  int[] differenceData;
  int[][] movementGraphData;
  int[][] movementGraphDataAvg;
  int movementTotal = 0;
  
  PImage preBlurImg;
  PImage resultImg;
  PImage historyMask;
  
  QBody currentBody;


  QData(int numBodies) {
    differenceData = new int[numBodies];
    movementGraphData = new int[numBodies][];
    movementGraphDataAvg = new int[numBodies][];
    for(int i=0; i < numBodies; ++i) {
      movementGraphData[i] = new int[MOVIE_WIDTH / numBodies];
      movementGraphDataAvg[i] = new int[MOVIE_WIDTH / numBodies];
    }
    resultImg = createImage(movie.width, movie.height, ARGB);
    historyMask = createImage(movie.width, movie.height, ARGB);
  }

  public int[] getDifferenceData() { return differenceData; }

  public int[][] getMovementGraphData() { return movementGraphData; }

  public int[][] getMovementGraphDataAvg() { return movementGraphDataAvg; }
  
  public int getMovementTotal() { return movementTotal; }

  public int computeFrameData() {
    int Xstart = 0;
    int Xend = 0;
    int Ystart = 0;
    int Yend = movie.height - 1;
    int total = 0;

    resultImg.loadPixels();
    if( !hideBackground ){
      preBlurImg = movie.get();
    }
    movie.filter(BLUR, 1);

    for(int i = 0; i < bodyStack.size(); ++i){
      currentBody = (QBody)bodyStack.get(i);
      Xstart = currentBody.L;
      Xend = currentBody.R;
      if( Xstart >= 0 ) differenceData[i] = getMovementSum(Xstart, Xend, Ystart, Yend);
      else {        // special case for the body that wraps around the panoramic video
        differenceData[i] = getMovementSum(0, Xend, Ystart, Yend);
        differenceData[i] += getMovementSum(movie.width + Xstart, movie.width-1, Ystart, Yend);
      }
      total += differenceData[i];
    }
    movementTotal = total;
    
    resultImg.updatePixels();
//    image(resultImg, 0, 0);    // update the display window with the framedifferenced image
//    arrayCopy(resultImg.pixels, pixels);
    if( movie.frame() > (analysisStartFrame + skipFirstFrames) )
      if( showMovie ) historyMask.blend(resultImg, 0, 0, historyMask.width, historyMask.height, 0, 0, resultImg.width, resultImg.height, ADD);
    if( showMovie ){
      if( showHistory )
        arrayCopy(historyMask.pixels, pixels);
      else
        arrayCopy(resultImg.pixels, pixels);
    }
    return total;
  }
  
  private int getMovementSum(int Xstart, int Xend, int Ystart, int Yend) {
    int movementSum = 0;     // Amount of movement in the frame
//      arraycopy(blankPixels, pixels);
      
    for(int y = Ystart; y <= Yend; ++y ){
      for(int x = Xstart; x <= Xend; ++x ){
        color currColor = movie.pixels[y*movie.width + x];
        color prevColor = previousFrame[y*movie.width + x];
        // Extract the red, green, and blue components from current pixel
        int currR = (currColor >> 16) & 0xFF; // Like red(), but faster
        int currG = (currColor >> 8) & 0xFF;
        int currB = currColor & 0xFF;
        // Extract red, green, and blue components from previous pixel
        int prevR = (prevColor >> 16) & 0xFF;
        int prevG = (prevColor >> 8) & 0xFF;
        int prevB = prevColor & 0xFF;
        // Compute the difference of the red, green, and blue values
        int diffR = abs(currR - prevR);
        int diffG = abs(currG - prevG);
        int diffB = abs(currB - prevB);
  
        // Sum the differences
        if (diffR + diffG + diffB > ui.minDiffThreshold) {   
          movementSum += 1;  // If this pixel's difference is greater than a minimum threshold, add to the running total
        }
        if( showMovie ) resultImg.pixels[y*movie.width + x] = getResultColor(diffR, diffG, diffB, x, y);          // Render the difference image to the screen
        previousFrame[y*movie.width + x] = currColor;           // Save the current color into the 'previous' buffer
      }
    }
    // if more than a certain percentage of the frame is different
//    if( (movementSum / NUM_PIXELS) > ui.minDiffSumThreshold) {
    if( (movementSum) > ui.minDiffSumThreshold) {
      return movementSum;
    }
    else
      return 0;
  }
  
  color getResultColor(int diffR, int diffG, int diffB, int x, int y) {
    color resultColor = color(0,0,0);
    int diffTotal = diffR + diffG + diffB;
    
    if ( diffTotal > ui.minDiffThreshold) {
      switch( ui.colorPalette ){
        case(0):
          // heat map coloring
          int heatIndex = constrain( (int)map( diffTotal, ui.minDiffThreshold, 255, 0, heatmap.width - 1 ), 0, heatmap.width - 1);
          resultColor = heatmap.get( heatIndex, heatIndex );
          break;
        case(1):
          // bright green difference
          resultColor = color(0,255,0);
          break;
        case(2):
          // true color difference
          resultColor = 0xff000000 | (diffR << 16) | (diffG << 8) | diffB;
          break;
      }
    }
    else {
      if( hideBackground)
        resultColor = color(0,0,0);  // show black
      else {
        color c = preBlurImg.get(x,y);
        resultColor = c;
//        resultColor = color( red(c)/2, green(c)/2, blue(c)/2 );  // show movie pixel at half brightness
      }
    }
    
    return resultColor;
  }
  
}

