/**
 * QMovement
 * by Daniel Steinbock <steinbock@stanford.edu>
 * based on Frame Differencing by Golan Levin. 
 * 
 * Quantify the amount of bodily movement in the movie using frame-differencing.
 *
 */ 

import processing.video.*;
import codeanticode.gsvideo.*;
import controlP5.*;
import toxi.math.*;


QUI ui;  // User interface
ControlP5 uiControl; // GUI control object (passed to QUI)

// Movie
GSMovie movie;
QMovieController mvc;
String videoFilename;

// Movie Capture
MovieMaker mm;  // Declare MovieMaker object


int[] blankPixels;
final static int MOVIE_WIDTH = 1152;
static long movieLength;

// Data I/O
static int NUM_DATA_COLUMNS = 2;


// Data Graph
QData qdata;
QGraph graph;
QOutput output;
static int movementSumMax = 0;

boolean hideBackground = true;
boolean showMovie = true;        // set display = false for batch processing
boolean showGraph = true;
boolean pauseDrawing = false;
boolean showHistory = false;
boolean showHighlight = false;
boolean showBodyBounds = false;

static int currFrame = 3;
long currTime = 0;
long prevTime = 0;

static boolean movieInitFlag = true;
static boolean setupBodyMode = false;

PImage[] selectionLoop;
PGraphics selectionBuffer;
static boolean initSelectionLoop = true;
static boolean initStartEndTimes = true;
static int loopFrame = 0;
static int loopFramesNum = 10;
static int loopFramesStepSize = 0;

static float analysisStartTime = 0.0;
static float analysisEndTime = 0.0;
static int analysisStartFrame = 0;
static int analysisEndFrame = 0;
static int skipFirstFrames = 10;   // # of initial frames to skip as analysis comes online

Stack delimiterStack;
Stack bodyStack;

PImage heatmap;
String path;
final static boolean RECORD_ON = true;
static String captureFilename;
static PImage captureImage;
static int captureStepSize = 2;
final static int MAX_DATA_VALUE = 500000;
final static float NUM_PIXELS = 1152*480;

void setup() {
  size(1152, 480);
  // Uses the default video input, see the reference if this causes an error
  frameRate(90);
  background(30);
  
  uiControl = new ControlP5(this);
  ui = new QUI(uiControl);
  delimiterStack = new Stack();
  bodyStack = new Stack();
  // Create MovieMaker object with size, filename,
  // compression codec and quality, framerate
  if( RECORD_ON ) {
//    mm = new MovieMaker(this, width, height, "Quaker-FrameDifferencing.mov", 5, MovieMaker.RAW, MovieMaker.LOSSLESS);
    captureImage = createImage(width, height, RGB);
  }
  textFont(loadFont("LucidaGrande-12.vlw"), 12.0);
  heatmap = loadImage("heatmap.jpg");
  // select path and load video
  path = selectInput("Select a video file ...");
  movie = new GSMovie(this, path);
  videoFilename = path.substring(path.lastIndexOf('/')+1, path.lastIndexOf('.'));
  mvc = new QMovieController(movie);
  movie.loop();
  selectionBuffer = createGraphics(1152, 320, JAVA2D);
}

void movieEvent(GSMovie m) {
}


void draw() {
    // When using video to manipulate the screen, use movie.available() and
    // movie.read() inside the draw() method so that it's safe to draw to the screen
  if (movie.available()) {
    if( movieInitFlag ){
      mvc.initMovie();
      movieInitFlag = false;
      setupBodyMode = true;
      ui.setStatus("Click to separate bodies.");
    }
    if( setupBodyMode ){    // column set up mode
      delimitBodies();
    }
    else {    // data analysis mode
      analyzeMovementData();
    }
  } // end if (movie.available())
}

void delimitBodies() {
  if(initStartEndTimes) {
    background(30);
    fill(255,255,255);
    textSize(16);
    text("Enter start and end times for analysis (in seconds). Enter 0 for no crop:", 100, 140);
  }
  else if(initSelectionLoop) {
    mov.jump( loopFrame * loopFramesStepSize + analysisStartFrame );
    mov.read();
    selectionBuffer.beginDraw();
    selectionBuffer.image(movie, 0, 0, mov.width, mov.height);    // draw movie
    selectionBuffer.fill(255,255,255);
    selectionBuffer.text((movie.frame() - analysisStartFrame) + " / " + (analysisEndFrame  - analysisStartFrame - 1), width-200, movie.height-2);
    selectionLoop[loopFrame] = selectionBuffer.get(0, 0, movie.width, movie.height);
    selectionBuffer.endDraw();
    background(30);
    fill(255,255,255);
    textSize(32);
    text("Loading movie... " + (movie.frame() - analysisStartFrame) + " / " + (analysisEndFrame  - analysisStartFrame - 1), width/2-200, movie.height/2);
    textSize(12);
    if( ++loopFrame >= selectionLoop.length ) {
      loopFrame = 0;
      initSelectionLoop = false;
    }
  }
  else {
    // loop through movie fast, playing every 1000
    background(30,0,0);
    image(selectionLoop[loopFrame], 0, 0);    // draw frames from the selection Loop
    stroke(100,200,240);  
    for(int i=0; i<delimiterStack.size(); ++i){
      line(((Integer)delimiterStack.get(i)).floatValue(), 0, ((Integer)delimiterStack.get(i)).floatValue(), movie.height-1);
    }
    if( ++loopFrame >= selectionLoop.length ) loopFrame = 0;
  }
}

void analyzeMovementData() {
  if( movie.frame() >= analysisEndFrame - 1 ) quit();
  movie.pause();
  movie.read();
  ++currFrame;
  mov.loadPixels(); // Make its pixels[] array available
  int movementTotal = qdata.computeFrameData();      // Amount of movement in the frame
  if( RECORD_ON && (movie.frame() > (analysisStartFrame + skipFirstFrames)) && (movie.frame() % captureStepSize == 0) ) {
//    mm.addFrame();  // Add window's pixels to movie
      captureFilename = path.substring(0,path.length()-4) + "-" + movie.frame() + ".jpg";
      println("Outputting " + captureFilename);
      captureImage = get(0,0,movie.width-1,437);
      captureImage.save(captureFilename);
  }
  currTime = System.currentTimeMillis();
  if( currFrame > (analysisStartFrame + 3) ) output.writeData();
  if( !pauseDrawing) {
    updatePixels();
    if(showBodyBounds)
      drawBodies();
    if( showGraph ) graph.draw();
    fill(200);
    text(movie.frame() + " / " + (analysisEndFrame - 1) + " | msec/frame = " + (currTime - prevTime) + " | fps = " + round(frameRate), width-350, height-2);
//    text(movie.frame() + " / " + (movieLength - 1) + " | " + movie.time() + " / " + movie.duration(), width-200, height-2);
  } else {
     println(movie.frame() + " / " + (analysisEndFrame - 1) + " | msec/frame = " + (currTime - prevTime));
  }
  prevTime = currTime;
  movie.play();
}

void saveBodies() {
  // iterate through the delimiterStack, putting them into a sortable vector
  Collections.sort(delimiterStack);
  int L = 0;
  int R = 0;
  for(int i=0; i<delimiterStack.size(); ++i){
    R = ((Integer)delimiterStack.get(i)).intValue();
    bodyStack.push(new QBody(L, R));
    L = R + 1;
  }
  if( delimiterStack.size() < 2 )
    bodyStack.push(new QBody(L, movie.width - 1));    // right-most QBody goes to video edge
  else
    ((QBody)bodyStack.get(0)).L = L - movie.width;          // first QBody gets negative L value to denote wrap-around from the right side
  

  // create all the data structures now that we know how many bodies there are
  qdata = new QData(bodyStack.size());
  graph = new QGraph();
  output = new QOutput(bodyStack.size());

  movie.speed(1.0);
  movie.play();
  movie.jump(analysisStartFrame);
  currFrame = analysisStartFrame;
  setupBodyMode = false;
  println("Processing " + bodyStack.size() + " bodies...");
}

void drawBodies() {
  for(int i=0; i<bodyStack.size(); ++i){
    ((QBody)bodyStack.get(i)).draw();
  }
}

void quit() {
  println("Finishing. Flushing. Exiting...");
  output.finish();
  if( RECORD_ON ) mm.finish();
  exit(); // Stop the program  
}


/**
 * GUI METHODS
 *
 */
 
void controlEvent(ControlEvent theEvent) {
//  // check if the event came from a group, if not, the event
//  // was triggered by a controller (go to else)
//  if(theEvent.isGroup()) {
//    if(theEvent.group() instanceof RadioButton) {
//      ui.setColorPalette(theEvent.group().value());
//    }
//  } 
//  else {
//    // if the event was not triggered by a group, 
//    // then we got an event from a controller.
    ui.dispatch(theEvent);
//  }

    
}

void mousePressed() {
  if( mouseY < movie.height && !initSelectionLoop){
    delimiterStack.push(new Integer(mouseX));
  }
}

void mouseReleased() {
}

void keyPressed() {
  if (key == ' ') {  // Finish the capture movie if space bar is pressed!
    mm.finish();  
  }
  if (key == 'u') {  // delete last delimiter
    if( !delimiterStack.empty() && setupBodyMode) {
      delimiterStack.pop();
      ui.setStatus("Last delimiter removed.");
    }
  }
  if (key == 's') { 
    ui.setStatus("Analysis in progress.");
    saveBodies();
  }
  if (key == 'c') {  // Reset the movement data
//    movementGraphData[0] = new int[1152];
  }
  if (key == 'p') {  // Reset the movement data
//    if( pauseDrawing ) frame.show();
//    else frame.hide();
    pauseDrawing = !pauseDrawing;
  }
  if (key == 'q') {  // Quit and flush output buffer
    quit();
  }
}


